package ru.dbotthepony.mc.otm.menu

import net.minecraft.network.FriendlyByteBuf
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.Slot
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.block.entity.matter.MatterPanelBlockEntity
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphListener
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.MMenus
import java.util.*
import java.util.function.Consumer
import java.util.function.Supplier

class CancelTaskPacket(val id: UUID) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeUUID(id)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val sender = context.sender!!
			(sender.containerMenu as? MatterPanelMenu)?.receiveTaskCancel(sender, id)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): CancelTaskPacket {
			return CancelTaskPacket(buff.readUUID())
		}
	}
}

class PatternsChangePacket(val isUpdate: Boolean, val patterns: Collection<IPatternState>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(isUpdate)
		buff.writeInt(patterns.size)

		for (pattern in patterns) {
			pattern.write(buff)
		}
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val menu = minecraft.player?.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			if (isUpdate) {
				menu.networkPatternsUpdated(patterns)
			} else {
				menu.networkPatternsRemoved(patterns)
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): PatternsChangePacket {
			val isUpdate = buff.readBoolean()
			val size = buff.readInt()

			return PatternsChangePacket(isUpdate, ArrayList<PatternState>(size).also {
				for (i in 0 until size) it.add(PatternState.read(buff) ?: throw NullPointerException("Unable to read PatternState from network at $i")) })
		}
	}
}

class TasksChangePacket(val isUpdate: Boolean, val tasks: Collection<IReplicationTask<*>>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeBoolean(isUpdate)
		buff.writeInt(tasks.size)

		for (task in tasks) {
			task.write(buff)
		}
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val menu = minecraft.player?.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			if (isUpdate) {
				menu.networkTasksUpdated(tasks)
			} else {
				menu.networkTasksRemoved(tasks)
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): TasksChangePacket {
			val isUpdate = buff.readBoolean()
			val size = buff.readInt()

			return TasksChangePacket(isUpdate, ArrayList<IReplicationTask<*>>(size).also {
				for (i in 0 until size) it.add(buff.readReplicationTask() ?: throw NullPointerException("Unable to read Replication Task from network at $i")) })
		}
	}
}

class ReplicationRequestPacket(val id: UUID, amount: Int) : MatteryPacket {
	val amount = amount.coerceAtLeast(1).coerceAtMost(99999)

	override fun write(buff: FriendlyByteBuf) {
		buff.writeUUID(id)
		buff.writeInt(amount)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val sender = context.sender ?: return@enqueueWork
			val menu = sender.containerMenu as? MatterPanelMenu ?: return@enqueueWork

			menu.requestReplication(sender, id, amount)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): ReplicationRequestPacket {
			return ReplicationRequestPacket(buff.readUUID(), buff.readInt())
		}
	}
}

class MatterPanelMenu @JvmOverloads constructor(
	p_38852_: Int,
	inventory: Inventory,
	tile: MatterPanelBlockEntity? = null
) : MatteryMenu(MMenus.MATTER_PANEL, p_38852_, inventory, tile), IMatterGraphListener {
	fun taskUpdated(task: IReplicationTask<*>) {
		sendNetwork(TasksChangePacket(true, listOf(task)))
	}

	fun taskRemoved(task: IReplicationTask<*>) {
		sendNetwork(TasksChangePacket(false, listOf(task)))
	}

	// client code
	val patterns = ArrayList<IPatternState>()
	val tasks = ArrayList<IReplicationTask<*>>()
	var changeset = 0

	fun networkPatternsUpdated(patterns: Collection<IPatternState>) {
		changeset++

		for (pattern in patterns) {
			val index = this.patterns.indexOfFirst(pattern::matchId)

			if (index != -1) {
				this.patterns[index] = pattern
			} else {
				this.patterns.add(pattern)
			}
		}
	}

	fun networkPatternsRemoved(patterns: Collection<IPatternState>) {
		changeset++

		for (pattern in patterns) {
			this.patterns.remove(pattern)
		}
	}

	fun networkTasksUpdated(tasks: Collection<IReplicationTask<*>>) {
		changeset++

		for (task in tasks) {
			val index = this.tasks.indexOfFirst(task::matchId)

			if (index != -1) {
				this.tasks[index] = task
			} else {
				this.tasks.add(task)
			}

			updateWatcher?.accept(task)
		}
	}

	fun networkTasksRemoved(tasks: Collection<IReplicationTask<*>>) {
		changeset++

		for (task in tasks) {
			this.tasks.remove(task)
			deleteWatcher?.accept(task)
		}
	}

	private var deleteWatcher: Consumer<IReplicationTask<*>>? = null
	private var updateWatcher: Consumer<IReplicationTask<*>>? = null

	fun networkTaskWatcher(updateWatcher: Consumer<IReplicationTask<*>>, deleteWatcher: Consumer<IReplicationTask<*>>) {
		this.deleteWatcher = deleteWatcher
		this.updateWatcher = updateWatcher
	}

	// server code
	fun requestReplication(ply: ServerPlayer, id: UUID, count: Int) {
		if (ply.isSpectator) {
			return
		}

		val tile = tile as MatterPanelBlockEntity? ?: return

		val graph = tile.matterGraph ?: return
		val state = graph.getPattern(id)

		if (state == null) {
			LOGGER.error("Received replication request from {} of {}, but it is not found in grid", ply, id)
			return
		}

		tile.addTask(state, count)
	}

	fun receiveTaskCancel(ply: ServerPlayer, id: UUID) {
		if (!ply.isSpectator)
			(tile as MatterPanelBlockEntity?)?.removeTask(id)
	}

	fun requestTaskCancel(id: UUID) {
		if (minecraft.player?.isSpectator != true)
			MenuNetworkChannel.sendToServer(CancelTaskPacket(id))
	}

	override fun onPatternAdded(state: IPatternState) {
		sendNetwork(PatternsChangePacket(true, listOf(state)))
	}

	override fun onPatternRemoved(state: IPatternState) {
		sendNetwork(PatternsChangePacket(false, listOf(state)))
	}

	override fun onPatternUpdated(newState: IPatternState, oldState: IPatternState) {
		sendNetwork(PatternsChangePacket(true, listOf(newState)))
	}

	private var initialSend = false

	private var listeningGrid: MatterNetworkGraph? = null

	init {
		if (tile != null) {
			listeningGrid = tile.matterGraph
			tile.attachMenu(this)
			listeningGrid?.addListener(this)
		}
	}

	override fun removed(p_38940_: Player) {
		super.removed(p_38940_)
		(tile as MatterPanelBlockEntity?)?.deatachMenu(this)
		listeningGrid?.removeListener(this)
	}

	override fun broadcastChanges() {
		super.broadcastChanges()
		if (!initialSend) fullPatternBroadcast()
	}

	private fun sendNetwork(packet: Any) {
		MenuNetworkChannel.send(PacketDistributor.PLAYER.with { inventory.player as ServerPlayer }, packet)
	}

	fun fullPatternBroadcast() {
		if (inventory.player.level.isClientSide) {
			initialSend = true
			return
		}

		val tile = tile as MatterPanelBlockEntity?

		if (tile != null) {
			val grid = tile.matterGraph

			if (grid != null) {
				initialSend = true
				sendNetwork(PatternsChangePacket(true, grid.patterns.toList()))
			}

			sendNetwork(TasksChangePacket(true, tile.allReplicationTasks.toList()))
		}
	}

	override val storageSlots: Collection<Slot>
		get() = emptyList()

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
